"""File containing main 2048 UI class and rendering logic

Author: Jakub Wlodek
Created: Nov-7-2019
"""

# We will use py_cui for the user interface
import py_cui

# Python utility libraries
import random
import copy

# The game module manages the game instance
import py_cui_2048.game as GAME


class CUI2048:
    """CUI 2048 class.

    Attributes
    ----------
    master : py_cui.PyCUI
        Main parent CUI object
    positions : list of py_cui.widgets.Button
        Buttons used to represent 2048 cells
    prev_board : list of list of int
        2D array representing previous board.
    prev_score : int
        Previous turns score
    undid_move : bool
        True if undo has just been used
    score_label : py_cui.widgets.Label
        label printing score and turn
    menu : py_cui.widgets.ScrollMenu
        Main CUI menu

    Methods
    -------
    update_turns_scores()
        Function that updates the score and turn counters as appropriate
    shift_up()
        Function that performs a move up (W) operation
    shift_down()
        Function that performs a move down (S) operation
    shift_left()
        Function that performs a move left (A) operation
    shift_right()
        Function that performs a move right (D) operation
    initialize_new_game()
        Function that creates a new game board
    operate_on_menu_item()
        Function that operates on the current selected menu item
    undo_move()
        Function that is used to undo/redo a move
    get_logo_text()
        Function tath gets ascii art 2048 logo
    apply_board_state()
        Function that applies the board state to the CUI.
    play_again()
        Function that asks user if they want to play again
    generate_initial_placement()
        Function that creates an initial board placement
    """

    def __init__(self, master):
        """Constructor for main 2048 CUI window object
        """

        self.master = master
        self.master.add_key_command(py_cui.keys.KEY_W_LOWER, self.shift_up)
        self.master.add_key_command(py_cui.keys.KEY_A_LOWER, self.shift_left)
        self.master.add_key_command(py_cui.keys.KEY_S_LOWER, self.shift_down)
        self.master.add_key_command(py_cui.keys.KEY_D_LOWER, self.shift_right)
        self.master.set_status_bar_text('Use WASD to move the board, Enter to enter the menu, and q to quit')
        
        self.positions = []
        self.prev_board = None
        self.prev_score = 0
        self.undid_move = False
        self.score_label = self.master.add_label('Score: 0, Turns: 0', 1, 4, column_span=3)
        for i in range (0,4):
            row = []
            for j in range(0,4):
                pos = self.master.add_button('0', i, j, command=None)
                row.append(pos)
            self.positions.append(row)
        self.master.add_block_label(self.get_logo_text(), 0, 4, row_span = 1, column_span=3)
        scroll_menu_options = ['New Game', 'Undo Move', 'Redo Move', 'Exit']
        self.menu = self.master.add_scroll_menu('Menu', 2, 4, row_span = 2, column_span = 3)
        self.menu.add_item_list(scroll_menu_options)
        self.menu.add_key_command(py_cui.keys.KEY_ENTER, self.operate_on_menu_item)
        self.menu.add_key_command(py_cui.keys.KEY_W_LOWER, self.shift_up)
        self.menu.add_key_command(py_cui.keys.KEY_A_LOWER, self.shift_left)
        self.menu.add_key_command(py_cui.keys.KEY_S_LOWER, self.shift_down)
        self.menu.add_key_command(py_cui.keys.KEY_D_LOWER, self.shift_right)
        self.menu.set_focus_text('Use WASD to move the board, arrows to select menu items, and enter to use a menu item')
        self.initialize_new_game()


    def update_turns_scores(self):
        """Function that updates the score and turn counters as appropriate
        """

        self.score_label.title = 'Score: {}, Turn: {}'.format(self.game_instance.score, self.game_instance.turn)


    def shift_up(self):
        """Function that performs a move up (W) operation
        """

        self.undid_move = False
        self.prev_board = copy.deepcopy(self.game_instance.game_board.board_positions)
        self.prev_score = self.game_instance.score
        valid = self.game_instance.game_board.process_columns('up')
        if valid:
            out = self.game_instance.game_board.add_random_tile()
            if out:
                self.apply_board_state()


    def shift_left(self):
        """Function that performs a move left (A) operation
        """

        self.undid_move = False
        self.prev_board = copy.deepcopy(self.game_instance.game_board.board_positions)
        self.prev_score = self.game_instance.score
        valid = self.game_instance.game_board.process_rows('left')
        if valid:
            out = self.game_instance.game_board.add_random_tile()
            if out:
                self.apply_board_state()

    def shift_down(self):
        """Function that performs a move down (S) operation
        """

        self.undid_move = False
        self.prev_board = copy.deepcopy(self.game_instance.game_board.board_positions)
        self.prev_score = self.game_instance.score
        valid = self.game_instance.game_board.process_columns('down')
        if valid:
            out = self.game_instance.game_board.add_random_tile()
            if out:
                self.apply_board_state()


    def shift_right(self):
        """Function that performs a move right (D) operation
        """

        self.undid_move = False
        self.prev_board = copy.deepcopy(self.game_instance.game_board.board_positions)
        self.prev_score = self.game_instance.score
        valid = self.game_instance.game_board.process_rows('right')
        if valid:
            out = self.game_instance.game_board.add_random_tile()
            if out:
                self.apply_board_state()


    def initialize_new_game(self):
        """Function that creates a new game board
        """

        self.prev_board = None
        self.prev_score = 0
        self.undid_move = False
        initial_placement = self.generate_initial_placement()
        self.game_instance = GAME.Game(initial_placement)
        self.apply_board_state()
        self.master.move_focus(self.menu)


    def operate_on_menu_item(self):
        """Function that operates on the current selected menu item
        """

        operation = self.menu.get()
        if operation == 'New Game':
            self.initialize_new_game()
        elif operation == 'Undo Move':
            self.undo_move(True)
        elif operation == 'Redo Move':
            self.undo_move(False)
        elif operation == 'Exit':
            exit()

    
    def undo_move(self, turn_up_down):
        """Function that is used to undo/redo a move

        Parameters
        ----------
        turn_up_down : bool
            Flag that specifies whether to perform undo or redo
        """

        if self.undid_move and turn_up_down:
            self.master.show_warning_popup('Warning', 'Only one Undo is allowed at a time!')
            return
        elif not self.undid_move and not turn_up_down:
            self.master.show_warning_popup('Warning', 'Cannot redo move that has not been undone!')
            return
        if self.prev_board is not None:
            temp_board = copy.deepcopy(self.game_instance.game_board.board_positions)
            self.game_instance.game_board.board_positions = copy.deepcopy(self.prev_board)
            self.prev_board = temp_board
            if turn_up_down:
                self.game_instance.turn = self.game_instance.turn - 1
                self.undid_move = True
            else:
                self.game_instance.turn = self.game_instance.turn + 1
                self.undid_move = False
            temp_score = self.game_instance.score
            self.game_instance.score = self.prev_score
            self.prev_score = temp_score
            self.apply_board_state()
        else:
            self.master.show_error_popup('Error', 'No move has been made!')


    def get_logo_text(self):
        """Function tath gets ascii art 2048 logo

        Returns
        -------
        out : str
            Ascii-art 2048 logo
        """

        out = ""
        out = out + " _____  _____    ___  _____ \n"
        out = out + "/ __  \|  _  |  /   ||  _  |\n"
        out = out + "`' / /'| |/' | / /| | \ V / \n"
        out = out + "  / /  |  /| |/ /_| | / _ \ \n"
        out = out + "./ /___\ |_/ /\___  || |_| |\n"
        out = out + "\_____/ \___/     |_/\_____/\n"
        return out


    def apply_board_state(self):
        """Function that applies the board state to the CUI.
        
        Certain values are assigned certain py_cui colors, and each cells value is written from the 
        game board
        """

        for i in range(0, 4):
            for j in range(0, 4):
                val = self.game_instance.game_board.board_positions[i][j]
                self.positions[i][j].title = str(val)
                if val == 0:
                    self.positions[i][j].color = py_cui.WHITE_ON_BLACK
                elif val == 2:
                    self.positions[i][j].color = py_cui.CYAN_ON_BLACK
                elif val == 4:
                    self.positions[i][j].color = py_cui.CYAN_ON_BLACK
                elif val == 8:
                    self.positions[i][j].color = py_cui.MAGENTA_ON_BLACK
                elif val == 16:
                    self.positions[i][j].color = py_cui.MAGENTA_ON_BLACK
                elif val == 32:
                    self.positions[i][j].color = py_cui.GREEN_ON_BLACK
                elif val == 64:
                    self.positions[i][j].color = py_cui.GREEN_ON_BLACK
                elif val == 128:
                    self.positions[i][j].color = py_cui.RED_ON_BLACK
                elif val == 256:
                    self.positions[i][j].color = py_cui.RED_ON_BLACK
                elif val == 512:
                    self.positions[i][j].color = py_cui.YELLOW_ON_BLACK
                elif val == 1024:
                    self.positions[i][j].color = py_cui.YELLOW_ON_BLACK
                elif val == 2048:
                    self.positions[i][j].color = py_cui.BLUE_ON_BLACK

        self.update_turns_scores()
        won = self.game_instance.check_victory()
        lost = self.game_instance.check_defeat()
        if won:
            self.master.show_yes_no_popup('Congratulations You Won! ! Play Again?', self.play_again)
        elif lost:
            self.master.show_yes_no_popup('Game Over, Score: {}. Play Again?'.format(self.game_instance.score), command=self.play_again)


    def play_again(self, response):
        """Function that asks user if they want to play again

        Parameters
        ----------
        response : bool
            User response to play again query
        """

        if response:
            self.initialize_new_game()
        else:
            exit()


    def generate_initial_placement(self):
        """Function that creates an initial board placement

        Returns
        -------
        initial_placement : list of list of int
            Two random coordinates for the initial values
        """

        initial_placement = []
        placement_1 = []
        placement_2 = []

        for i in range(0,2):
            placement_1.append(random.randint(0,3))
            placement_2.append(random.randint(0,3))

        placement_1.append(random.randint(1,2) * 2)
        placement_2.append(random.randint(1,2) * 2)

        initial_placement.append(placement_1)
        initial_placement.append(placement_2)
        return initial_placement


def main():
    """Program entrypoint, create CUI, wrapper object and start it
    """

    root = py_cui.PyCUI(4,7)
    root.set_title('2048')
    cui_2048 = CUI2048(root)
    root.start()


from setuptools import setup, find_packages
from sys import platform
from setuptools.command.install import install
import os

with open('requirements.txt', 'r') as req_fp:
    required_packages = req_fp.readlines()

# Use README for long description
with open('README.md', 'r') as readme_fp:
    long_description = readme_fp.read()

setup(
    name="py_cui_2048",
    version="0.0.1",
    author="Jakub Wlodek",
    author_email="jwlodek.dev@gmail.com",
    description="A command line version of the classic 2048 game, built with py_cui.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="BSD 3-Clause",
    keywords="tui command-line cli cui curses 2048",
    url="https://github.com/jwlodek/py_cui_2048",
    packages = find_packages(exclude=['tests', 'docs']),
    entry_points={
        'console_scripts': [
            'py2048 = py_cui_2048:main',
        ],
    },
    install_requires=required_packages,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

)


"""File containing game board logic and functions (written independantly of CUI)

Author: Jakub Wlodek
Created: Nov-12-2019
"""

# Use random lib for random cell placement
import random


class Game:
    """Game instance class for 2048

    Attributes
    ----------
    game_board : py_cui_2048.game.Board
        The game board object representation
    score : int
        The current game score
    turn : int
        The current turn counter

    Methods
    -------
    check_victory()
        Checks if user won
    check_defeat()
        Checks if no valid moves left
    """

    def __init__(self, initial_placement):
        """Constructor for game instace class object
        """

        self.game_board = Board(self, initial_placement)
        self.score = 0
        self.turn = 0


    def check_victory(self):
        """Checks if user won

        Returns
        -------
        won : bool
            True if 2048 tile found, false otherwise
        """

        for row in self.game_board.board_positions:
            for elem in row:
                if elem == 2048:
                    return True


    def check_defeat(self):
        """Checks if no valid moves left

        Either two neighbors are the same, or there is an empty location, otherwise
        no legal moves left

        Returns
        -------
        defeat : bool
            True if no moves left, otherwise false
        """

        for i in range(0, len(self.game_board.board_positions)):
            for j in range(0, len(self.game_board.board_positions[i])):
                if j > 0:
                    if self.game_board.board_positions[i][j] == self.game_board.board_positions[i][j-1]:
                        return False
                if i > 0:
                    if self.game_board.board_positions[i][j] == self.game_board.board_positions[i-1][j]:
                        return False
                if self.game_board.board_positions[i][j] == 0:
                    return False

        return True


class Board:
    """Class representing the game board itself, and operations on it

    Attributes
    ----------
    parent_instance : py_cui_2048.game.Game
        Parent game instance object
    board_positions : list of list of int
        2D array containing board values

    Methods
    -------
    get_empty_pos_list()
        Gets list of empty positions
    add_random_tile()
        Adds a new random tile (either 2 or 4) into an empty location after valid user move
    process_columns()
        Processes a vertical (up/down) movement
    process_rows()
        Processes a horizontal (left/right) movement
    update_nums()
        Top level update function called from 2048 client
    print_board()
        Debug function for printing board positions
    """

    def __init__(self, parent_instance, initial_placement):
        """Constructor for board class
        """

        self.parent_instance = parent_instance
        initial_x1 = initial_placement[0][0]
        initial_y1 = initial_placement[0][1]
        initial_val1 = initial_placement[0][2]
        initial_x2 = initial_placement[1][0]
        initial_y2 = initial_placement[1][1]
        initial_val2 = initial_placement[1][2]
        self.board_positions = [[0,0,0,0],[0,0,0,0],[0,0,0,0],[0,0,0,0]]
        self.board_positions[initial_y1][initial_x1] = initial_val1
        self.board_positions[initial_y2][initial_x2] = initial_val2


    def get_empty_pos_list(self):
        """Function for getting list of empty cells in board

        Returns
        -------
        empty_pos_list : list of list of int
            list of coordinates of empty cells
        """

        empty_pos_list = []
        for i in range(0, len(self.board_positions)):
            for j in range(0, len(self.board_positions[i])):
                if self.board_positions[i][j] == 0:
                    empty_pos_list.append([i,j])
        return empty_pos_list


    def add_random_tile(self):
        """Function that adds a new tile to a random empty location

        Returns
        -------
        added : bool
            True if success, false otherwise
        """

        try:
            possible_pos = self.get_empty_pos_list()
            rand_val = ((random.randint(1, 10) % 2) + 1) * 2
            rand_pos = possible_pos[random.randint(0, len(possible_pos) - 1)]
            self.board_positions[rand_pos[0]][rand_pos[1]] = rand_val
            return True
        except ValueError:
            return False


    def process_columns(self, direction):
        """Processes a vertical (up/down) movement

        Parameters
        ----------
        direction : str
            Either up or down
        
        Returns
        -------
        was_valid : bool
            True if board changed, false otherwise
        """

        was_valid = False
        for i in range(0, len(self.board_positions[0])):
            temp = []
            for j in range(0, len(self.board_positions)):
                temp.append(self.board_positions[j][i])
            if direction == 'up':
                temp.reverse()
            valid, score = self.update_nums(temp)
            if valid:
                self.parent_instance.score = self.parent_instance.score + score
                was_valid = True
            if direction == 'up':
                temp.reverse()
            for j in range(0, len(self.board_positions)):
                self.board_positions[j][i] = temp[j]

        if was_valid:
            self.parent_instance.turn = self.parent_instance.turn + 1
        return was_valid


    def process_rows(self, direction):
        """Processes a horizontal (left/right) movement

        Parameters
        ----------
        direction : str
            Either left or right
        
        Returns
        -------
        was_valid : bool
            True if board changed, false otherwise
        """

        was_valid = False
        for i in range(0, len(self.board_positions)):
            temp = self.board_positions[i]
            if direction == 'left':
                temp.reverse()
            valid, score = self.update_nums(temp)
            if valid:
                self.parent_instance.score = self.parent_instance.score + score
                was_valid = True
            if direction == 'left':
                temp.reverse()
            self.board_positions[i] = temp

        if was_valid:
            self.parent_instance.turn = self.parent_instance.turn + 1

        return was_valid


    def update_nums(self, nums):
        """Function that updates the numbers in a single column or row based on 2048 rules

        Always updates as if it is a row moving left to right. Caller functions convert to this format
        and convert back afterwords, so all use this function

        Parameters
        ----------
        nums : list of int
            The row or column to update
        
        Returns
        -------
        valid_move : bool
            True if row changed, false otherwise
        score_gain : int
            Amount of score the move gained
        """

        valid_move = False
        score_gain = 0
        counter = 3
        while counter > 0:
            next_num = counter - 1
            while  next_num > 0 and nums[next_num] == 0:
                next_num = next_num - 1
            if nums[counter] == nums[next_num]:
                nums[next_num] = 2 * nums[counter]
                valid_move = True
                score_gain = score_gain + (nums[counter] * 2)
                nums[counter] = 0
                counter = next_num
            counter = next_num
        
        output_arr = []
        counter = 3
        non_zero = 0
        while counter >= 0:
            if nums[counter] != 0:
                output_arr.append(nums[counter])
                non_zero = non_zero + 1
            counter = counter - 1
        for i in range(0, 4-non_zero):
            output_arr.append(0)
        for i in range(0, len(output_arr)):
            if nums[i] != output_arr[3-i]:
                valid_move = True
            nums[i] = output_arr[3-i]
        return valid_move, score_gain


    def print_board(self):
        """Debug function for printing board
        """
        
        for i in range(0, len(self.board_positions)):
            print(self.board_positions[i])